// main.cpp
// ========
// BMP Image Viewer
//
//  AUTHOR: Song Ho Ahn (song.ahn@gmail.com)
// CREATED: 2006-04-28
// UPDATED: 2006-05-12
///////////////////////////////////////////////////////////////////////////////

#include <GL/glut.h>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <cstring>                      // for strcpy()
#include "Tga.h"                        // TGA loader

using std::stringstream;
using std::cout;
using std::endl;
using std::ends;



// CALLBACK functions for GLUT ////////////////////////////////////////////////
void displayCB();
void reshapeCB(int w, int h);
void keyboardCB(unsigned char key, int x, int y);

// function declarations //////////////////////////////////////////////////////
void initGL();
int  initGLUT(int argc, char **argv);
void drawString(const char *str, int x, int y, float color[4], void *font);
void showInfo();

// CONSTANTS //////////////////////////////////////////////////////////////////
const char *FILE_NAME = "lena_color.tga";
const int MAX_NAME = 512;               // max length of string

// global variables ///////////////////////////////////////////////////////////
char    fileName[MAX_NAME];             // image file name
int     imageWidth, imageHeight;        // image resolution
int     mainWin;                        // GLUT window ID
void    *font = GLUT_BITMAP_8_BY_13;    // GLUT font type
GLenum  format;                         // pixel format (GL_LUMINANCE, GL_RGB, GL_BGR_EXT, ...)
GLenum  type;                           // data type (GL_UNSIGNED_BYTE, GL_INT, ...)
const unsigned char *inBuf;             // image buffer
Image::Tga tga;                         // TGA loader object



///////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
    // use default image file if not specified
    if(argc == 2)
    {
        strcpy(fileName, argv[1]);
    }
    else{
        cout << "Usage: " << argv[0] << " <image-file>\n";
        strcpy(fileName, FILE_NAME);
        cout << "\nUse default image \"" << fileName << "\"" << endl;
    }

    // load TGA image
    if(!tga.read(fileName))
    {
        cout << "ERROR: " << tga.getError() << endl;
        return 0;     // exit if failed load image
    }

    // copy bmp infos
    imageWidth = tga.getWidth();
    imageHeight = tga.getHeight();
    inBuf = tga.getDataRGB();
    type = GL_UNSIGNED_BYTE;    // only allow BMP with 8-bit per channel

    // We assume the image is 8-bit, 24-bit or 32-bit BMP
    int bpp = tga.getBitCount();
    if(bpp == 8)
        format = GL_LUMINANCE;
    else if(bpp == 24)
        format = GL_RGB;
    else if(bpp == 32)
        format = GL_RGBA;
    else
        return 0;               // NOT supported, exit

    // init GL and GLUT
    initGLUT(argc, argv);
    initGL();

    // the last GLUT call (LOOP) ///////////////////////////////////////////////
    // window will be shown and display callback is triggered by events
    // NOTE: this call never return to main().
    glutMainLoop(); /* Start GLUT event-processing loop */

    return 0;
} // END OF MAIN //////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
// initialize GLUT for windowing
///////////////////////////////////////////////////////////////////////////////
int initGLUT(int argc, char **argv)
{
    // GLUT stuff for windowing
    // initialization openGL window.
    // it is called before any other GLUT routine
    glutInit(&argc, argv);

    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);   // display mode

    glutInitWindowSize(imageWidth, imageHeight);    // window size

    glutInitWindowPosition(100, 100);               // window location

    // finally, create a window with openGL context
    // Window will not displayed until glutMainLoop() is called
    // it returns a unique ID
    int handle = glutCreateWindow(argv[0]);     // param is the title of window

    // register GLUT callback functions
    glutDisplayFunc(displayCB);
    glutReshapeFunc(reshapeCB);
    glutKeyboardFunc(keyboardCB);
    //glutTimerFunc(100, timerCB, 100);             // redraw only every given millisec
    //glutMouseFunc(mouseCB);
    //glutMotionFunc(mouseMotionCB);

    return handle;
}



///////////////////////////////////////////////////////////////////////////////
// initialize OpenGL
// disable unused features
///////////////////////////////////////////////////////////////////////////////
void initGL()
{
    glClearColor(0, 0, 0, 0);
    glShadeModel(GL_FLAT);  // shading mathod: GL_SMOOTH or GL_FLAT
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

    // TUNNING IMAGING PERFORMANCE
    // turn off  all pixel path and per-fragment operation
    // which slows down OpenGL imaging operation (glDrawPixels glReadPixels...).
    glDisable(GL_ALPHA_TEST);
    glDisable(GL_BLEND);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_DITHER);
    glDisable(GL_FOG);
    glDisable(GL_LIGHTING);
    glDisable(GL_LOGIC_OP);
    glDisable(GL_STENCIL_TEST);
    glDisable(GL_TEXTURE_1D);
    glDisable(GL_TEXTURE_2D);

    glPixelTransferi(GL_MAP_COLOR, GL_FALSE);
    glPixelTransferi(GL_RED_SCALE, 1);
    glPixelTransferi(GL_RED_BIAS, 0);
    glPixelTransferi(GL_GREEN_SCALE, 1);
    glPixelTransferi(GL_GREEN_BIAS, 0);
    glPixelTransferi(GL_BLUE_SCALE, 1);
    glPixelTransferi(GL_BLUE_BIAS, 0);
    glPixelTransferi(GL_ALPHA_SCALE, 1);
    glPixelTransferi(GL_ALPHA_BIAS, 0);
    //glPixelZoom(1.0, 1.0);
}



///////////////////////////////////////////////////////////////////////////////
// write 2d text using GLUT
// The projection matrix must be set to orthogonal before call this function.
///////////////////////////////////////////////////////////////////////////////
void drawString(const char *str, int x, int y, float color[4], void *font)
{
    //glPushAttrib(GL_LIGHTING_BIT | GL_CURRENT_BIT); // lighting and color mask
    //glDisable(GL_LIGHTING);     // need to disable lighting for proper text color

    glColor4fv(color);          // set text color
    glRasterPos2i(x, y);        // place text position

    // loop all characters in the string
    while(*str)
    {
        glutBitmapCharacter(font, *str);
        ++str;
    }

    //glEnable(GL_LIGHTING);
    //glPopAttrib();
}



///////////////////////////////////////////////////////////////////////////////
// display info messages
///////////////////////////////////////////////////////////////////////////////
void showInfo()
{
    // backup current model-view matrix
    glPushMatrix();                     // save current modelview matrix
    glLoadIdentity();                   // reset modelview matrix

    // set to 2D orthogonal projection
    glMatrixMode(GL_PROJECTION);     // switch to projection matrix
    glPushMatrix();                  // save current projection matrix
    glLoadIdentity();                // reset projection matrix
    glOrtho(0, imageWidth, 0, imageHeight, -1, 1);// set to orthogonal projection

    float color[4] = {1, 1, 1, 1};

    stringstream ss;
    ss << std::fixed << std::setprecision(3);
    ss << "Name: " << fileName << ends;  // add 0(ends) at the end
    drawString(ss.str().c_str(), 1, imageHeight-14, color, font);
    ss.str(""); // clear buffer

    ss << "Width: " << imageWidth << ends;
    drawString(ss.str().c_str(), 1, imageHeight-(14*2), color, font);
    ss.str("");

    ss << "Height: " << imageHeight << ends;
    drawString(ss.str().c_str(), 1, imageHeight-(14*3), color, font);

    // unset floating format
    ss << std::resetiosflags(std::ios_base::fixed | std::ios_base::floatfield);

    // restore projection matrix
    glPopMatrix();                   // restore to previous projection matrix

    // restore modelview matrix
    glMatrixMode(GL_MODELVIEW);      // switch to modelview matrix
    glPopMatrix();                   // restore to previous modelview matrix
}



//=============================================================================
// CALLBACKS
//=============================================================================

void displayCB(void)
{
    //glutSetWindow(mainWin);
    glClear(GL_COLOR_BUFFER_BIT);       // clear canvas

    // specify current raster position(coordinate)
    glRasterPos2i(0, imageHeight);      // upper-left corner in openGL coordinates
    glPixelZoom(1.0f, -1.0f);           // swap the y-direction of image to top-to-bottom
                                        // Otherwise, the image will be up side down.

    // draw the image
    glDrawPixels(imageWidth, imageHeight, format, type, inBuf);
    glPixelZoom(1, 1); // back to normal orientation

    // draw some messages on the screen
    showInfo();

    //glFlush();
    glutSwapBuffers();
}



void reshapeCB(int w, int h)
{
    glViewport(0, 0, (GLsizei)w, (GLsizei)h);

    // main-window
    //glutSetWindow(mainWin);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0, imageWidth, 0, imageHeight, -1, 1);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}


void keyboardCB(unsigned char key, int x, int y)
{

    switch(key)
    {
    case 27: // ESCAPE
        exit(0);
        break;

    case 'r':
    case 'R':
        break;
    }
    glutSetWindow(mainWin);
    glutPostRedisplay();
}

